using dg1d
using CairoMakie
CairoMakie.activate!()


function analyze(dir)

  tomlfile = get_parameter_file(dir)
  parameters = FlatDict(TOML.parsefile(joinpath(dir, tomlfile)))

  mesh = make_Mesh(trim_headers(parameters, "Mesh"))

  id_prms = trim_headers(parameters, "Initialdata", "Mesh")
  id_fn = ScalarEquation.make_function_initialdata_equation(id_prms)
  # Using Initialdata and Mesh parameters is not ideal: Mesh.K could be different between
  # runs, but we know that make_function_initialdata_equation should not care about Mesh.K
  # Perhaps we need to redefine the meaning of the Mesh section: E.g Mesh refers to the dimensions
  # of the physical domain, whereas grid referes to the layout of the computationl domain?
  id_parameters = trim_headers(parameters, "Initialdata", "Mesh")
  id_fn = ScalarEquation.make_function_initialdata_equation(id_parameters)
  v = parameters["Equation.v"]
  analytic_u(x, t) = id_fn(periodize(x - v * t, mesh.L))

  thisdir = dirname(@__FILE__)
  output = Output(joinpath(thisdir, first(splitext(tomlfile)), "output.h5"))
  ts = output.times

  # select samples to plot
  nsamples = 5
  ntotal   = length(ts)
  nstep    = trunc(Int64, ntotal / (nsamples-1))
  rng      = collect(1:nstep:ntotal)
  plot_ts     = ts[rng]
  plot_us     = [ read_variable(output, "dynamic_variables.u", t) for t in plot_ts ]
  plot_ana_us = [ analytic_u.(view(mesh.x,:), t) for t in plot_ts ]
  
  # plotting
  fig = Figure(resolution=(1600,1200))
  colors = to_colormap(:rainbow, length(rng))
  # dim 'older' results
  transparency(idx, n) = exp((idx-n)*2/n)
  ax1 = Axis(fig[2,1], ylabel=L"u", xlabel=L"x")
  ax2 = Axis(fig[2,2], ylabel=L"u_{ana}", xlabel=L"x")
  ax3 = Axis(fig[2,3], ylabel=L"u - u_{ana}", xlabel=L"x")
  L = layout(mesh)
  Npts, K = L
  x_w_nans = make_nan_separated_storage(L)
  x_w_nans[1:end-1,:] .= mesh.x
  y_w_nans = make_nan_separated_storage(L)

  handles, labels = [], String[]
  for (i, (u, ana_u, t)) in enumerate(zip(plot_us, plot_ana_us, plot_ts))
    mat_u     = reshape(u, L)
    mat_ana_u = reshape(ana_u, L)

    transp = transparency(i, nsamples+1)

    y_w_nans[1:end-1,:] .= mat_u
    h = lines!(ax1, x_w_nans[:], y_w_nans[:], color=(colors[i],transp))
    push!(handles, h)
    push!(labels, "t=$t")

    y_w_nans[1:end-1,:] .= mat_ana_u
    lines!(ax2, x_w_nans[:], y_w_nans[:], color=(colors[i],transp))

    y_w_nans[1:end-1,:] .= mat_u .- mat_ana_u
    lines!(ax3, x_w_nans[:], y_w_nans[:], color=(colors[i],transp))
  end

  xlims!(ax1, extrema(mesh.x))
  xlims!(ax2, extrema(mesh.x))

  Legend(fig[1,1:2], handles, labels, orientation=:horizontal, tellwidth=true)

  save(joinpath(dir, "evolution.png"), fig)

  return
end


#######################################################################
#                                Main                                 #
#######################################################################


projectdir = abspath(dirname(@__FILE__))
dg1d.analyze_projects(analyze, projectdir)
