using dg1d
import Base: Filesystem
using GLMakie
GLMakie.activate!()


function readdata(dir)
  outputfile = joinpath(dir, "output.h5")
  us = Vector{Float64}[]
  ts = Float64[]
  x = h5open(outputfile,"r") do file
    groups = keys(file)
    filter!(g->g!="mesh", groups)
    groups = parse.(Int64, groups)
    sort!(groups)
    for g in groups
      push!(us, vec(read(file[string(g)]["dynamic_variables.u"])))
      push!(ts, read(file[string(g)]["t"]))
    end
    x = read(file["mesh"]["mesh"])
  end

  Npts, K = size(x)

  x = vec(x)

  return Npts, K, x, ts, us
end


function analyze(dir)

  thisdir = basename(@__DIR__)

  tomlfile = get_parameter_file(dir)
  parameters = FlatDict(TOML.parsefile(joinpath(dir, tomlfile)))

  mesh = make_Mesh(trim_headers(parameters, "Mesh"))

  equation = ScalarEquation.make_Equation(trim_headers(parameters, "Equation"))
  @unpack v = equation

  id_parameters = trim_headers(parameters, "Initialdata", "Mesh")
  id_fn = ScalarEquation.make_function_initialdata_equation(id_parameters)
  analytic_u(x, t) = id_fn(periodize(x - v * t, mesh.L))

  # read data
  Npts, K, x, ts, us = readdata(dir)

  tstart, tend = 0, ts[end]
  increment_t = tstart

  # select samples to plot
  ntotal = length(ts)
  @show ntotal
  max_time_steps = 1000
  indices, sample_indices = if ntotal < max_time_steps
    indices = collect(1:ntotal)
    indices, indices
  else
    index_increment = ntotal / max_time_steps
    sample_indices = [ trunc(Int64, floor((i-1) * index_increment+1)) for i = 1:max_time_steps ]
    sample_indices = unique(sample_indices)
    collect(1:max_time_steps), sample_indices
  end
  plot_ts = ts[sample_indices]
  plot_us = us[sample_indices]
  plot_ana_us = [ analytic_u.(x, t) for t in plot_ts ]
  

  time = Observable(plot_ts[1])

  # prepare plot
  fig = Figure(resolution=(1600,1200))
  colors = to_colormap(:rainbow)

  # plot
  Label(fig[1,1:2], text=@lift("t = $($time)"))
  subfig1 = fig[2,1]
  ax1 = Axis(subfig1, ylabel="u", xlabel="x")
  subfig2 = fig[2,2]
  ax2 = Axis(subfig2, ylabel="|u - u_ana|", xlabel="x")
  x = reshape(x, (Npts, K))

  # use arrays with NaNs in last row to separate data sets
  x_w_nans = zeros(Float64, Npts+1, K)
  x_w_nans[1:end-1, :] .= x
  x_w_nans[end, :] .= NaN
  x_w_nans = x_w_nans[:]
  u_w_nans = deepcopy(x_w_nans)
  err_w_nans = deepcopy(x_w_nans)

  # plot first data set
  shape = (Npts, K)
  nan_shape = (Npts+1, K)
  @views begin
    mat_u_w_nans = reshape(u_w_nans, nan_shape)
    mat_u_w_nans[1:end-1, :] .= reshape(plot_us[1], shape)
    mat_err_w_nans = reshape(err_w_nans, nan_shape)
    mat_err_w_nans[1:end-1, :] .= reshape(plot_ana_us[1], shape) .- mat_u_w_nans[1:end-1, :]
  end

  plt_u_w_nans = Observable(u_w_nans)
  plt_err_w_nans = Observable(err_w_nans)

  lines!(ax1, x_w_nans, plt_u_w_nans)
  lines!(ax2, x_w_nans, plt_err_w_nans)

  framerate = 30

  record(fig, joinpath(thisdir, "movie.mp4"), indices; framerate=framerate) do idx
    time[] = plot_ts[idx]
    mat_u_w_nans = reshape(plt_u_w_nans.val, nan_shape)
    mat_u_w_nans[1:end-1, :] .= reshape(plot_us[idx], shape)
    mat_err_w_nans = reshape(plt_err_w_nans.val, nan_shape)
    mat_err_w_nans[1:end-1, :] .= reshape(plot_ana_us[idx], shape) .- mat_u_w_nans[1:end-1, :]
    # trigger update
    plt_u_w_nans[] = plt_u_w_nans[]
    plt_err_w_nans[] = plt_err_w_nans[]
  end

  return
end


#######################################################################
#                                Main                                 #
#######################################################################


analyze("advection_sine/advection_sine_24")
