#######################################################################
#                             Parameters                              #
#######################################################################


"""
$(FUNCTIONNAME)
$(FIELDS)
"""
@option struct Parameters_Evolution <: dg1d.AbstractParameters
  "Time stepping algorithm, choose one out of LSERK, RK4, SSPRK3, ..."
  algorithm::String = "LSERK"
  "End time of evolution"
  tend::Float64 = 1
end


function sanitize(prms::Parameters_Evolution)
  @assert prms.algorithm in [ "LSERK", "RK4", "SSPRK3" ]
  @assert tstart < tend
end


#######################################################################
#                          Butcher-Tablaeus                           #
#######################################################################


"""
    BT_MIDPT

2nd order explicit Runge Kutta method -- midpoint formula.
"""
Base.@kwdef struct BT_MIDPT
  nstages::Int64 = 2
  a::Matrix{Float64} = [ 0.0     0.0;
                         0.5     0.0 ]
  b::Vector{Float64} = [ 0.0,    1.0 ]
  c::Vector{Float64} = [ 0.0,    0.5 ]
end


"""
    BT_RK4

4th order explicit Runge Kutta method.
"""
Base.@kwdef struct BT_RK4
  nstages::Int64 = 4
  a::Matrix{Float64} = [ 0.0     0.0     0.0     0.0;
                         0.5     0.0     0.0     0.0;
                         0.0     0.5     0.0     0.0;
                         0.0     0.0     1.0     0.0 ]
  b::Vector{Float64} = [ 1.0/6.0,  1.0/3.0,  1.0/3.0,  1.0/6.0 ]
  c::Vector{Float64} = [ 0.0,      0.5,      0.5,      1.0 ]
end


"""
    BT_RK4_twothirds

4th order explicit Runge Kutta method -- two thirds version.
"""
Base.@kwdef struct BT_RK4_twothirds
  nstages::Int64 = 4
  a::Matrix{Float64} = [ 0.0       0.0       0.0     0.0;
                         1.0/3.0   0.0       0.0     0.0;
                         -1.0/3.0  1.0       0.0     0.0;
                         1.0       -1.0      1.0     0.0 ]
  b::Vector{Float64} = [ 1.0/8.0,  3.0/8.0, 3.0/8.0, 1.0/8.0 ]
  c::Vector{Float64} = [ 0.0,      1.0/3.0, 2.0/3.0, 1.0 ]
end


"""
    BT_RK4_Ralston

4th order explicit Runge Kutta method -- Ralston's method,
designed to have minimal truncation error.
cf. https://en.wikipedia.org/wiki/List_of_Runge%E2%80%93Kutta_methods#Classic_fourth-order_method
"""
Base.@kwdef struct BT_RK4_Ralston
  nstages::Int64 = 4
  a::Matrix{Float64} = [ 0.0         0.0           0.0         0.0;
                         0.4         0.0           0.0         0.0;
                         0.29697761  0.15875964    0.0         0.0;
                         0.21810040  -3.05096516   3.83286476  0.0 ]
  b::Vector{Float64} = [ 0.17476028, -0.55148066,  1.20553560, 0.17118478  ]
  c::Vector{Float64} = [ 0.0,        0.4,          0.45573725, 1.0         ]
end


"""
    BT_SSPRK3

Strong stability preserving Runge Kutta 3rd order method.
"""
Base.@kwdef struct BT_SSPRK3
  nstages::Int64 = 3
  a::Matrix{Float64} = [ 0.0     0.0     0.0;
                         1.0     0.0     0.0;
                         0.25    0.25    0.0 ]
  b::Vector{Float64} = [ 1.0/6.0,  1.0/6.0,  2.0/3.0 ]
  c::Vector{Float64} = [ 0.0,      1.0,      0.5 ]
end


"""
    BT_RKF45

Runge-Kutta-Fehlberg 4/5 method,
cf. https://en.wikipedia.org/wiki/Runge%E2%80%93Kutta%E2%80%93Fehlberg_method
"""
Base.@kwdef struct BT_RKF45
  nstages::Int64 = 6
  a::Matrix{Float64} = [ 0           0           0           0           0       0;
                         1/4         0           0           0           0       0;
                         3/32        9/32        0           0           0       0;
                         1932/2197   -7200/2197  7296/2197   0           0       0;
                         439/216     -8          3680/513    -845/4104   0       0;
                         -8/27       2           -3544/2565  1859/4104   -11/40  0 ]
  b1::Vector{Float64} = [ 16/135     0           6656/12825  28561/56430 -9/50   2/55 ]
  b2::Vector{Float64} = [ 25/216     0           1408/2565   2197/4104   -1/5    0 ]
  c::Vector{Float64}  = [ 0          1/4         3/8         12/13       1       1/2 ]
end


"""
    BT_RKF10

Runge-Kutta-Feagin 10th order method.
Coefficients taken from https://codereview.stackexchange.com/q/83635
"""
struct BT_RKF10
  nstages::Int64
  a::Matrix{Float64}
  b::Vector{Float64}
  c::Vector{Float64}

  function BT_RKF10()
    nstages = 17
    a = zeros(Float64, nstages, nstages)
    b = zeros(Float64, nstages)
    c = zeros(Float64, nstages)

    a[02,01] = 1//10

    a[03,01] = -0.915176561375291440520015019275342154318951387664369720564660
    a[03,02] = 1.45453440217827322805250021715664459117622483736537873607016

    a[04,01] = 0.202259190301118170324681949205488413821477543637878380814562
    a[04,03] = 0.606777570903354510974045847616465241464432630913635142443687

    a[05,01] = 0.184024714708643575149100693471120664216774047979591417844635
    a[05,03] = 0.197966831227192369068141770510388793370637287463360401555746
    a[05,04] = -0.0729547847313632629185146671595558023015011608914382961421311

    a[06,01] = 0.0879007340206681337319777094132125475918886824944548534041378
    a[06,04] = 0.410459702520260645318174895920453426088035325902848695210406
    a[06,05] = 0.482713753678866489204726942976896106809132737721421333413261

    a[07,01] = 0.0859700504902460302188480225945808401411132615636600222593880
    a[07,04] = 0.330885963040722183948884057658753173648240154838402033448632
    a[07,05] = 0.489662957309450192844507011135898201178015478433790097210790
    a[07,06] = -0.0731856375070850736789057580558988816340355615025188195854775

    a[08,01] = 0.120930449125333720660378854927668953958938996999703678812621
    a[08,05] = 0.260124675758295622809007617838335174368108756484693361887839
    a[08,06] = 0.0325402621549091330158899334391231259332716675992700000776101
    a[08,07] = -0.0595780211817361001560122202563305121444953672762930724538856

    a[09,01] = 0.110854379580391483508936171010218441909425780168656559807038
    a[09,06] = -0.0605761488255005587620924953655516875526344415354339234619466
    a[09,07] = 0.321763705601778390100898799049878904081404368603077129251110
    a[09,08] = 0.510485725608063031577759012285123416744672137031752354067590

    a[10,01] = 0.112054414752879004829715002761802363003717611158172229329393
    a[10,06] = -0.144942775902865915672349828340980777181668499748506838876185
    a[10,07] = -0.333269719096256706589705211415746871709467423992115497968724
    a[10,08] = 0.499269229556880061353316843969978567860276816592673201240332
    a[10,09] = 0.509504608929686104236098690045386253986643232352989602185060

    a[11,01] = 0.113976783964185986138004186736901163890724752541486831640341
    a[11,06] = -0.0768813364203356938586214289120895270821349023390922987406384
    a[11,07] = 0.239527360324390649107711455271882373019741311201004119339563
    a[11,08] = 0.397774662368094639047830462488952104564716416343454639902613
    a[11,09] = 0.0107558956873607455550609147441477450257136782823280838547024
    a[11,10] = -0.327769124164018874147061087350233395378262992392394071906457

    a[12,01] = 0.0798314528280196046351426864486400322758737630423413945356284
    a[12,06] = -0.0520329686800603076514949887612959068721311443881683526937298
    a[12,07] = -0.0576954146168548881732784355283433509066159287152968723021864
    a[12,08] = 0.194781915712104164976306262147382871156142921354409364738090
    a[12,09] = 0.145384923188325069727524825977071194859203467568236523866582
    a[12,10] = -0.0782942710351670777553986729725692447252077047239160551335016
    a[12,11] = -0.114503299361098912184303164290554670970133218405658122674674

    a[13,01] = 0.985115610164857280120041500306517278413646677314195559520529
    a[13,04] = 0.330885963040722183948884057658753173648240154838402033448632
    a[13,05] = 0.489662957309450192844507011135898201178015478433790097210790
    a[13,06] = -1.37896486574843567582112720930751902353904327148559471526397
    a[13,07] = -0.861164195027635666673916999665534573351026060987427093314412
    a[13,08] = 5.78428813637537220022999785486578436006872789689499172601856
    a[13,09] = 3.28807761985103566890460615937314805477268252903342356581925
    a[13,10] = -2.38633905093136384013422325215527866148401465975954104585807
    a[13,11] = -3.25479342483643918654589367587788726747711504674780680269911
    a[13,12] = -2.16343541686422982353954211300054820889678036420109999154887

    a[14,01] = 0.895080295771632891049613132336585138148156279241561345991710
    a[14,03] = 0.197966831227192369068141770510388793370637287463360401555746
    a[14,04] = -0.0729547847313632629185146671595558023015011608914382961421311
    a[14,06] = -0.851236239662007619739049371445966793289359722875702227166105
    a[14,07] = 0.398320112318533301719718614174373643336480918103773904231856
    a[14,08] = 3.63937263181035606029412920047090044132027387893977804176229
    a[14,09] = 1.54822877039830322365301663075174564919981736348973496313065
    a[14,10] = -2.12221714704053716026062427460427261025318461146260124401561
    a[14,11] = -1.58350398545326172713384349625753212757269188934434237975291
    a[14,12] = -1.71561608285936264922031819751349098912615880827551992973034
    a[14,13] = -0.0244036405750127452135415444412216875465593598370910566069132

    a[15,01] = -0.915176561375291440520015019275342154318951387664369720564660
    a[15,02] = 1.45453440217827322805250021715664459117622483736537873607016
    a[15,05] = -0.777333643644968233538931228575302137803351053629547286334469
    a[15,07] = -0.0910895662155176069593203555807484200111889091770101799647985
    a[15,13] = 0.0910895662155176069593203555807484200111889091770101799647985
    a[15,14] = 0.777333643644968233538931228575302137803351053629547286334469

    a[16,01] = 1//10
    a[16,03] = -0.157178665799771163367058998273128921867183754126709419409654
    a[16,15] = 0.157178665799771163367058998273128921867183754126709419409654

    a[17,01] = 0.181781300700095283888472062582262379650443831463199521664945
    a[17,02] = 27//40
    a[17,03] = 0.342758159847189839942220553413850871742338734703958919937260
    a[17,05] = 0.259111214548322744512977076191767379267783684543182428778156
    a[17,06] = -0.358278966717952089048961276721979397739750634673268802484271
    a[17,07] = -1.04594895940883306095050068756409905131588123172378489286080
    a[17,08] = 0.930327845415626983292300564432428777137601651182965794680397
    a[17,09] = 1.77950959431708102446142106794824453926275743243327790536000
    a[17,10] = 1//10
    a[17,11] = -0.282547569539044081612477785222287276408489375976211189952877
    a[17,12] = -0.159327350119972549169261984373485859278031542127551931461821
    a[17,13] = -0.145515894647001510860991961081084111308650130578626404945571
    a[17,14] = -0.259111214548322744512977076191767379267783684543182428778156
    a[17,15] = -0.342758159847189839942220553413850871742338734703958919937260
    a[17,16] = -27//40

    b[01] = 1//30
    b[02] = 1//40
    b[03] = 1//30
    b[05] = 1//20
    b[07] = 1//25
    b[09] = 0.189237478148923490158306404106012326238162346948625830327194
    b[10] = 0.277429188517743176508360262560654340428504319718040836339472
    b[11] = 0.277429188517743176508360262560654340428504319718040836339472
    b[12] = 0.189237478148923490158306404106012326238162346948625830327194
    b[13] = -1//25
    b[14] = -1//20
    b[15] = -1//30
    b[16] = -1//40
    b[17] = 1//30

    c[02] = 1//10
    c[03] = 0.539357840802981787532485197881302436857273449701009015505500
    c[04] = 0.809036761204472681298727796821953655285910174551513523258250
    c[05] = 0.309036761204472681298727796821953655285910174551513523258250
    c[06] = 0.981074190219795268254879548310562080489056746118724882027805
    c[07] = 5//6
    c[08] = 0.354017365856802376329264185948796742115824053807373968324184
    c[09] = 0.882527661964732346425501486979669075182867844268052119663791
    c[10] = 0.642615758240322548157075497020439535959501736363212695909875
    c[11] = 0.357384241759677451842924502979560464040498263636787304090125
    c[12] = 0.117472338035267653574498513020330924817132155731947880336209
    c[13] = 5//6
    c[14] = 0.309036761204472681298727796821953655285910174551513523258250
    c[15] = 0.539357840802981787532485197881302436857273449701009015505500
    c[16] = 1//10
    c[17] = 1

    return new(nstages, a, b, c)
  end
end


"""
    COEFFS_LSERK4

Low storage explicit Runge Kutta 4th order method.
Taken from Hesthaven, Warburton (2007), section 3.4.
"""
Base.@kwdef struct COEFFS_LSERK4
  nstages::Int64 = 5
  a::Vector{Float64} = 
         [            0.0,
          -567301805773.0/1357537059087.0,
          -2404267990393.0/2016746695238.0,
          -3550918686646.0/2091501179385.0,
          -1275806237668.0/842570457699.0]
  b::Vector{Float64} = 
         [ 1432997174477.0/9575080441755.0,
           5161836677717.0/13612068292357.0,
           1720146321549.0/2090206949498.0,
           3134564353537.0/4481467310338.0,
           2277821191437.0/14882151754819.0]
  c::Vector{Float64} = 
         [             0.0,
           1432997174477.0/9575080441755.0,
           2526269341429.0/6820363962896.0,
           2006345519317.0/3224310063776.0,
           2802321613138.0/2924317926251.0]
end


#######################################################################
#                              Steppers                               #
#######################################################################


"""
    LSERK4_step!(u!, F, u0, t, dt, tmp_k, coeffs, callback_substep)

Stepper for Low Storage five-stage fourth-order Explicit Runge Kutta scheme.
Taken from Hesthaven, Warburton (2007), section 3.4.
- `u!` - mutable RHS of ODE
- `F` - RHS functional
- `u0` - initial data
- `t` - current time
- `dt` - time step
- `coeffs` - Explicit Runge Kutta coefficients
"""
function LSERK4_step!(u!, F, u0, t, dt, tmp_k, coeffs, callback_substep)

  @unpack nstages, a, b, c = coeffs

  kim1, ki = tmp_k
  N = length(u!)

  for i = 1:nstages

    ti = t + c[i] * dt
    if i == 1
      F(ki, u0, ti)
    else
      F(ki, u!, ti)
    end

    @. ki *= dt
    if i == 1
      @. u! = u0 + b[i] * ki
    else
      ai = a[i]
      @. ki += ai * kim1
      @. u! += b[i] * ki
    end

    success = callback_substep(u!, ti, i) # +1 because u! is value for next stage
    !success && return false

    if i < nstages
      @. kim1 = ki
    end

  end # for

  return true
end


"""
    baseERK_step!(u!, F, u0, t, dt, tmp_k, coeffs, callback_substep)

Base stepper for Explicit Runge Kutta scheme.
- `u!` - mutable RHS of ODE
- `F` - RHS functional
- `u0` - initial data
- `t` - current time
- `dt` - time step
- `coeffs` - Explicit Runge Kutta coefficients, e.g. Butcher tableau
"""
function baseERK_step!(u!, F, u0, t, dt, tmp_k, coeffs, callback_substep)
  @unpack nstages, a, b, c = coeffs

  tn = t
  for n = 1:nstages
    tn = t + c[n] * dt
    u! .= u0
    for i = 1:n-1
      u! .+= a[n, i] * tmp_k[i]
    end
    if n > 1
      success = callback_substep(u!, tn, n-1)
      !success && return false
    end
    F(tmp_k[n], u!, tn)
    tmp_k[n] .*= dt
  end

  u! .= u0
  for n = 1:nstages
    u! .+= b[n] * tmp_k[n]
  end

  success = callback_substep(u!, tn, nstages)

  return true
end


"""
    make_RK_stages(size, n)

Generate `n` stages of length `size`. To be used with with `baseERK_step` and `LSERK4_step` 
as temporary storage for substeps.
"""
function make_RK_stages(size, n)
  return [ zeros(Float64, size) for _ in 1:n ]
end


"""
    make_ERK_stepper(u0, scheme::Symbol, callback_substep)

Generate an Explicit-Runge-Kutta stepper for iterate `u0` that uses scheme `scheme`.
"""
function make_ERK_stepper(u0, scheme::Symbol, callback_substep)

  coeffs, method = if scheme == :MIDPT
    BT_MIDPT(), baseERK_step!
  elseif  scheme == :RK4
    BT_RK4(), baseERK_step!
  elseif  scheme == :RK4_twothirds
    BT_RK4_twothirds(), baseERK_step!
  elseif  scheme == :RK4_Ralston
    BT_RK4_Ralston(), baseERK_step!
  elseif  scheme == :SSPRK3
    BT_SSPRK3(), baseERK_step!
  elseif  scheme == :LSERK4
    COEFFS_LSERK4(), LSERK4_step!
  elseif  scheme == :RKF10
    BT_RKF10(), baseERK_step!
  else
    error("Requested ERK scheme not implemented!")
  end

  tmp_k = make_RK_stages(size(u0), coeffs.nstages)

  stepper(u!, F, u0, t, dt) = method(u!, F, u0, t, dt, tmp_k, coeffs, callback_substep)

  return stepper
end


"""
    isinvalid(u)

Return false if any component of `u` is either `NaN` or `Inf`.
"""
function isinvalid(u)
  for ui in u
    ( isnan(ui) || isinf(ui) ) && return true
  end
  return false
end


"""
    evolve_ERK(F, u0, tstep::Union{Function,Real}, tspan::Tuple{Real,Real}, scheme::Symbol;
               callback_fullstep::Union{Function,Nothing}=nothing,
               callback_substep::Union{Function,Nothing}=nothing)

Use an Explicit-Runge-Kutta scheme to evolve a function `F` from initial state `u0`
with time step `tstep` for times in `tspan=(tstart,tend)` using the scheme `scheme`.

Provide a method for `tstep(u,t,n_attempts)::Float64` to evolve with a state and time
dependent time step.
`u, t` are the current iterate and time, `n_attempts` is the number of attempts made
to retake the current time step. The latter is only needed to adjust the step in case
a substep failed, see `callback_substep` below.
Return `nothing` to terminate the evolution.

Provide `callback_fullstep(u, t, iter)` to run this function after every full step. Any return
value is ignored.

Provide `callback_substep(u, t, n_stage)::Bool` to run this function after every sub step.
The return value indicates success. Use `tstep(u,t,n_attempts)` to retry the full step
with a smaller step.
"""
function evolve_ERK(
    F, u0, tstep::Function, tspan::Tuple{Real,Real}, scheme::Symbol;
    callback_fullstep=nothing, callback_substep=nothing)

  # enforce interface
  @assert length(methods(tstep, [Any,Real,Int])) > 0 """
  `tstep` must take arguments `(u,t,n_attempts)`.
  """
  cb_fullstep = if isnothing(callback_fullstep) ||
      (callback_fullstep isa CallbackSet && length(callback_fullstep) == 0)
    (u,t,n) -> true
  else
    @assert length(methods(callback_fullstep, [Any,Real,Int])) > 0 """
    `callback_fullstep` must take arguments `(u,t,n_attempts)`.
    """
    callback_fullstep
  end
  cb_substep = if isnothing(callback_substep) ||
      (callback_substep isa CallbackSet && length(callback_substep) == 0)
    (u,t,n) -> true
  else
    @assert length(methods(callback_substep, [Any,Real,Int])) > 0 """
    `callback_substep` must take arguments `(u,t,n_attempts)`.
    """
    callback_substep
  end

  # setup storage
  un = deepcopy(u0)
  unp1 = zeros(Float64, length(un))
  step! = make_ERK_stepper(u0, scheme, cb_substep)

  tstart, tend = tspan
  tn = tstart
  prog = Progress(100) # 100 %
  iter = 0

  # invoke once before evolution to
  cb_fullstep(un, tn, iter, force=true)

  n_retake_attempts = 0
  dt = -1

  while tn < tend

    prcnt = tn / (tend-tstart) * 100
    ProgressMeter.update!(prog, trunc(Int64,prcnt))

    new_dt = if tstep isa Function
      new_dt = tstep(un, tn, n_retake_attempts)
      if isnothing(new_dt)
        println()
        @info """Termination requested by `tstep`!"""
        unp1 .= un # restore previous state
        break
      end
      new_dt
    else
      tstep
    end

    # shorten time step to hit tend exactly
    if tn + new_dt > tend
      new_dt = tend - tn
      # in this case it might happen that retaking did not reduce the step width enough
      # we then just try again
      if n_retake_attempts > 0 && new_dt >= dt
        n_retake_attempts += 1
        continue
      end
    end

    # is step admissible?
    if n_retake_attempts > 0 && new_dt >= dt
      println()
      @warn """Attempted to retake a step with too large of a step width:
               (dt, proposed dt) = ($dt, $new_dt)
               Terminating evolution ..."""
      unp1 .= un # restore previous state
      break
    end

    dt = new_dt

    # step now
    success = step!(unp1, F, un, tn, dt)
    if !success
      n_retake_attempts += 1
      continue
    else
      n_retake_attempts = 0
    end

    if isinvalid(unp1)
      println()
      @warn """NaN or Inf encountered in evolution at t = $(tn)!
                Terminating evolution ..."""
      unp1 .= un # restore previous state
      break
    end

    un .= unp1
    tn += dt
    iter += 1

    cb_fullstep(un, tn, iter)

    if tn >= tend
      ProgressMeter.finish!(prog)
    end
  end # while

  return unp1
end


function evolve_ERK(F, u0, tstep::Real, tspan::Tuple{Real,Real}, scheme::Symbol; kwargs...)
  return evolve_ERK(F, u0, (u,t,n)->tstep, tspan, scheme; kwargs...)
end


@doc """
    evolve_adaptive_ERK(F, u0, tstep, tspan; 
                        reltol=1e-10, abstol=1e-10, 
                        scheme::Symbol=:RK4, post_hook=nothing)

Similar to `evolve_ERK`, but iterates every ERK step till the difference between two
consecutive steps obeys the following criterion:
```
# un, unp1 are the results from the nth and (n+1)th iteration
norm(un - unp1, 1) <= reltol * norm(un, 1) + abstol
```
"""
function evolve_adaptive_ERK(F, u0, tstep, tspan; reltol=1e-10, abstol=1e-10,
    scheme::Symbol=:RK4, post_hook=nothing)

  un = deepcopy(u0); unp1 = similar(un);
  tmp_u1 = similar(unp1); tmp_u2 = similar(unp1);
  cb_substep = (u,t,n) -> true
  stepper! = make_ERK_stepper(u0, scheme, cb_substep)

  adaptive_max_its = 20 # maximum number of adaptive iterations

  tstart, tend = tspan
  tn = tstart
  prog = Progress(100) # 100 %
  while tn < tend

    prcnt = tn / (tend-tstart) * 100
    ProgressMeter.update!(prog, trunc(Int,prcnt))

    if isa(tstep, Function)
      dt = tstep(un, tn)
    else
      dt = tstep
    end

    if tn + dt > tend
      dt = tend - tn
    end

    adapted_dt = dt
    norm_un = norm(un, 1)
    # estimate error by difference between a single RK step and
    # two consecutive RK step evolutions
    # iterate this until tolerance is reached
    adapted_it = 1
    while true
      stepper!(unp1,    F, un,      tn,                 adapted_dt)
      stepper!(tmp_u1,  F, un,      tn,                 adapted_dt*0.5)
      stepper!(tmp_u2,  F, tmp_u1,  tn+adapted_dt*0.5,  adapted_dt*0.5)

      if isinvalid(unp1) || isinvalid(tmp_u1) || isinvalid(tmp_u2)
        println()
        error("NaN or Inf encountered in evolution at t = $(tn)! Aborting ...")
      end

      @. tmp_u1 = unp1 - tmp_u2 # reuse tmp_u1 storage for error computation
      err = norm(tmp_u1, 1)
      err <= reltol * norm_un + abstol && break

      adapted_dt *= 0.5
      adapted_it += 1
      if adapted_it > adaptive_max_its
      # if log(dt/adapted_dt)/log(2) > adaptive_max_its
        println()
        error("Could not determine solution after $adaptive_max_its refinement iterations! " *
              "Aborting ... ")
      end
    end

    un .= unp1
    tn += adapted_dt

    if tn >= tend
      ProgressMeter.finish!(prog)
    end

    if !isnothing(post_hook)
      success::Bool = post_hook(un, tn)
      if !success
        println()
        @info "Termination requested by post_hook!"
        break
      end
    end

  end # while

  return unp1
end
