export FlatDict,
       filter_headers,
       trim_headers,
       prepend_headers


#######################################################################
#                                Types                                #
#######################################################################


"""
    FlatDict([itr]) <: AbstractDict{String,V}
    FlatDict([itr], header) <: AbstractDict{String,V}

Dictionary for handling flattened parameter files.
Add the `header` argument to prepend a header to all keys.
Offers convinience methods like `trim_headers, filter_headers, add_header`.
"""
struct FlatDict{V} <: AbstractDict{String,V}
  dict::Dict{String,V}
  function FlatDict(d::AbstractDict, header="")
    flat_d = flatten(d)
    if length(header) > 0
      flat_d = prepend_keys(flat_d, header)
    end
    return new{valtype(flat_d)}( Dict( (String(k), v) for (k,v) in pairs(flat_d) ) )
  end
end


#######################################################################
#                            Constructors                             #
#######################################################################


FlatDict(args...) = FlatDict(Dict(args...))
FlatDict{V}(args...) where {V} = FlatDict(Dict{String,V}(args...))


#######################################################################
#                            Base methods                             #
#######################################################################


Base.get(flatd::FlatDict, key, default) = get(flatd.dict, key, default)
Base.length(flatd::FlatDict) = length(flatd.dict)
Base.getindex(flatd::FlatDict, elements...) = getindex(flatd.dict, elements...)
Base.setindex!(flatd::FlatDict, value, elements...) = setindex!(flatd.dict, value, elements...)
Base.iterate(flatd::FlatDict) = iterate(flatd.dict)
Base.iterate(flatd::FlatDict, index) = iterate(flatd.dict, index)
Base.merge(flatd::FlatDict, others::AbstractDict...) = FlatDict(merge(flatd.dict, others...))
Base.pop!(flatd::FlatDict, key, default) = pop!(flatd.dict, key, default)


#######################################################################
#                               Methods                               #
#######################################################################


function _filter_header(flatd::FlatDict, header)
  rgx = Regex("^$(string(header))\\.")
  matched_dict, _ = match_keys(k -> match(rgx, string(k)), flatd.dict) 
  return FlatDict(matched_dict)
end


"""
    filter_headers(flatd::FlatDict, headers...)

Returns a subset of the dictionary `flatd` for which they keys match one of `headers`.
"""
function filter_headers(flatd::FlatDict, headers...)
  # display(flatd)
  # for header in headers
  #   display(header)
  #   _filter_header(flatd, header)
  # end
  sub_flatds = [ _filter_header(flatd, header) for header in headers ]
  return merge(sub_flatds...)
end


function _trim_header(flatd::FlatDict, header)
  if header[end] != '.'
    header *= "."
  end
  trimmed_dict = trim_keys(flatd.dict, string(header))
  return FlatDict(trimmed_dict)
end


"""
    trim_headers(flatd::FlatDict, headers...)

Returns a subset of the dictionary `flatd` for which they keys matched one of `headers` and
also removed `headers` from the key.
"""
function trim_headers(flatd::FlatDict, headers...)
  # cannot precompute with filter_headers(flatd, headers...), because _trim_header(flatd, header)
  # only trims one header at a time, but the filtered result would still contain multiple headeres
  trimmed_flatds = [ _trim_header(filter_headers(flatd, header), header) for header in headers ]
  return merge(trimmed_flatds...)
end


function prepend_headers(flatd::FlatDict, headers...)
  for header in reverse(headers)
    flatd = prepend_keys(flatd, header)
  end
  return flatd
end
