module LGL


using Jacobi, LinearAlgebra
import dg1d: lagrange_poly, first_derivative_matrix, purge_zeros!



@doc """
Quadrature nodes and weights for Gauss-Legendre rule.
n_pts ... number of quadrature points
Exact up to order 2*n_pts - 1.
"""
function rule(n_pts)
  α = 0.0; β = 0.0
  x = zglj(n_pts, α, β)
  w = wglj(x, α, β)
  return x, w
end



@doc """
Inner product with Legendre weight w(x) = 1.
x ... Vector of LGL nodes
w ... Vector of LGL weights
f ... Vector with function values evaluated at nodes x.
"""
function inner_product(x, w, f)
  @assert length(w) == length(x)
  @assert length(w) == length(f)
  sum = dot(w, f)
  return sum
end 



@doc """
Inverse of exact full mass matrix, computed using Sherman-Morrison formula,
cf. PhD thesis Marcus Bugner, Notes spectral elements Bernd Bruegmann,
Kopriva (2013), Teukolsky (2015).
n_pts ... number of qudrature nodes = number of lagrange polynomials
"""
function inverse_mass_matrix_exact(n_pts)
  z, w = rule(n_pts)
  N = n_pts - 1
  invM = zeros(n_pts, n_pts)
  P = legendre.(z, N)
  invM = diagm(1.0 ./ w)
  invM += 0.5 * n_pts * P * transpose(P)
  return invM
end 



@doc """
n_pts ... number of qudrature nodes = number of lagrange polynomials
"""
function inverse_mass_matrix_lumped(n_pts)
  _, w = rule(n_pts)
  invM = diagm(1.0 ./ w)
  return invM
end 



@doc """
M_ij = (li, lj)
n_pts ... number of qudrature nodes = number of lagrange polynomials
"""
function mass_matrix_exact(n_pts)
  z, w = rule(n_pts)
  M = zeros(n_pts, n_pts)
  aux_z, aux_w = rule(n_pts+2)
  L(i, x) = lagrange_poly(i, x, z)
  M = zeros(n_pts, n_pts)
  for i=1:Npts, j=1:Npts
    li = L.(i, z)
    for j = i:n_pts
      lj = L.(j, z)
      M[i,j] = inner_product(aux_z, aux_w, li .* lj)
      if (i != j)
        M[j,i] = M[i,j]
      end
    end
  end # for

  return M
end 



@doc """
M_ij = (li, lj)
Approximated mass matrix due to under integration of LGL quadrature.
n_pts ... number of qudrature nodes = number of lagrange polynomials
"""
function mass_matrix_lumped(n_pts)
  z, w = rule(n_pts)
  M = diagm(w)
  return M
end 



@doc """
S_ij = (li, dlj/dx)
Exactly representable, because product is of order 2N - 1 which can be integrated
exactly using LGL quadrature.
n_pts ... number of quadrature nodes = number of lagrange polynomials
"""
function stiffness_matrix(n_pts)
  z, w = rule(n_pts)
  D = first_derivative_matrix(z)
  S = w .* D   
  return S
end 



@doc """
L_ij = (P_i, P_j)
Analogon to mass matrix of Lagrange polynomials, expresses the linear dependency
of the numerical representation of the Legendre basis. 
x ... GLGL nodes
w ... GLGL weights
"""
function linear_dependency_matrix_lumped(x::tV, w::tV) where {tV}
  N = length(x)
  L = zeros(N,N)
  for i in 1:N
    Pi = map(x->legendre(x, i-1), x)
    for j in i:N
      Pj = map(x->legendre(x, j-1), x)
      lij = inner_product(x, w, Pi .* Pj)
      L[i,j] = lij
      if i != j
        L[j,i] = L[i,j]
      end
    end
  end
  purge_zeros!(L)

  return L
end


@doc """
E_ij = (l_i, P_j)
Commonly used to project either Lagrange polynomials onto Legendres or
reverse.
x ... GLGL nodes
w ... GLGL weights
"""
function expansion_matrix_lumped(x, w)
  N = length(x)
  T = zeros(N,N)
  for i in 1:N
    li = map(xi->lagrange_poly(i, xi, x), x)
    for j in 1:N
      Pj = map(x->legendre(x, j-1), x)
      tij = inner_product(x, w, li .* Pj)
      T[i,j] = tij
    end
  end
  purge_zeros!(T)

  return T
end


@doc """
L_ij = (P_i, P_j)
Analogon to mass matrix of Lagrange polynomials, expresses the linear dependency
of the numerical representation of the Legendre basis. 
x ... GLGL nodes
w ... GLGL weights
"""
function linear_dependency_matrix_exact(x, w)
  N = length(x)
  N2 = N+2
  x2, w2 = rule(N2)
  L = zeros(N,N)
  for i in 1:N
    Pi = map(x->legendre(x, i-1), x2)
    for j in i:N
      Pj = map(x->legendre(x, j-1), x2)
      lij = inner_product(x2, w2, Pi .* Pj)
      L[i,j] = lij
      if i != j
        L[j,i] = L[i,j]
      end
    end
  end
  purge_zeros!(L)

  return L
end


@doc """
E_ij = (l_i, P_j)
Commonly used to project either Lagrange polynomials onto Legendres or
reverse.
x ... GLGL nodes
w ... GLGL weights
"""
function expansion_matrix_exact(x, w)
  N = length(x)
  N2 = N + 2
  x2, w2 = rule(N2)
  T = zeros(N,N)
  for i in 1:N
    li = map(xi->lagrange_poly(i, xi, x), x2)
    for j in 1:N
      Pj = map(x->legendre(x, j-1), x2)
      tij = inner_product(x2, w2, li .* Pj)
      T[i,j] = tij
    end
  end
  purge_zeros!(T)

  return T
end


end # module LGL
