function make_callback(rhs::RHS, bdryconds::BoundaryConditions)
  cbfn_equation(u, t) = callback_equation(u, t, rhs, isperiodic(bdryconds), rhs.equation)
  cb_equation         = FunctionCallback(cbfn_equation,
                                         CallbackTiming(every_iteration=1,every_dt=0))
  cbfn_tci(u, t)      = callback_tci(u, t, rhs, isperiodic(bdryconds), rhs.tci)
  cb_tci              = FunctionCallback(cbfn_tci,
                                         CallbackTiming(every_iteration=1, every_dt=0))
  cbfn_hrsc(u, t)     = callback_hrsc(u, t, rhs, isperiodic(bdryconds), rhs.hrsc)
  cb_hrsc             = FunctionCallback(cbfn_hrsc,
                                         CallbackTiming(every_iteration=1,every_dt=0))
  callbackset = CallbackSet(cb_equation, cb_tci, cb_hrsc)
  return callbackset
end


#######################################################################
#                              Equation                               #
#######################################################################


function callback_equation(state_u, state_t, rhs, isperiodic, eq)
  @unpack cache = rhs
  wrap_dynamic_variables!(cache, state_u)
  @unpack t, tm1 = get_global_variables(cache)
  tm1 .= t
  t   .= state_t
  # compute_entropy_variables!(rhs)
end


#######################################################################
#                                HRSC                                 #
#######################################################################


callback_hrsc(state_u, state_t, rhs, isperiodic, hrsc::Nothing) = nothing
callback_hrsc(state_u, state_t, rhs, isperiodic, hrsc::AbstractReconstruction) = nothing


function callback_hrsc(state_u, state_t, rhs, isperiodic, hrsc::Maybe{HRSC.AbstractHRSC}) 
  TODO(callback_hrsc)
end


function callback_hrsc(state_u, state_t, rhs, isperiodic, hrsc::HRSC.AbstractArtificialViscosity)
  @unpack cache, mesh, equation = rhs
  @unpack max_v                 = get_static_variables(cache)
  @unpack mu, cellmax_v         = get_cell_variables(cache)
  @unpack sD                    = get_dynamic_variables(cache)
  broadcast_volume!(speed, equation, cache)
  Npts, K = layout(mesh)
  mat_cellmax_v = reshape(view(max_v, :), (Npts, K))
  for k = 1:K
    cellmax_v[k] = maximum(view(mat_cellmax_v, :, k))
  end
  HRSC.compute_viscosity!(
      mu,
      sD, cellmax_v,
      hrsc)
end


function callback_hrsc(state_u, state_t, rhs, isperiodic, hrsc::HRSC.TCIViscosity)
  @unpack cache, mesh = rhs
  @unpack mu, flag    = get_cell_variables(cache)
  HRSC.compute_viscosity!(mu, flag, hrsc)
end


function callback_hrsc(state_u, state_t, rhs, isperiodic, hrsc::HRSC.SmoothedArtificialViscosity)
  callback_hrsc(state_u, state_t, rhs, isperiodic, hrsc.av)
  @unpack cache       = rhs
  @unpack mu          = get_cell_variables(cache)
  indices = findall(muk -> muk > 0.0, mu)
  tmp = zeros(size(mu))
  _, K = layout(rhs.mesh)
  # artificially broaden viscosity
  for k in indices
    k_lhs, k_rhs = periodic_index(k-1, K), periodic_index(k+1, K)
    muk = mu[k]
    tmp[k_lhs] = max(muk, tmp[k_lhs])
    tmp[k] = max(muk, tmp[k])
    tmp[k_rhs] = max(muk, tmp[k_rhs])
  end
  @. mu = tmp
  @unpack smoothed_mu = get_static_variables(cache)
  smoothed_mu .= vec(hrsc.smoother(mu, isperiodic))
end


function callback_hrsc(state_u, state_t, rhs, isperiodic, hrsc::HRSC.EntropyViscosity)
  TODO(callback_hrsc)
  @unpack cache, mesh            = rhs
  @unpack rho                    = get_dynamic_variables(cache)
  @unpack S, Sm1, flx_S, flx_Sm1 = get_static_variables(cache)
  @unpack mu, cellmax_v          = get_cell_variables(cache)
  @unpack t, tm1                 = get_global_variables(cache)
  HRSC.compute_viscosity!(
      mu,
      rho, cellmax_v, S, Sm1, flx_S, flx_Sm1, t[1], tm1[1],
      hrsc)
end


#######################################################################
#                                 TCI                                 #
#######################################################################


function callback_tci(state_u, state_t, rhs, isperiodic, tci::Nothing) end


function callback_tci(state_u, state_t, rhs, isperiodic, tci::TCI.Threshold)
  error("Threshold TCI must be combined with an indicator that is supposed to be zero for " *
        "for smooth solutions, but not for shocks. TODO Implement an indicator variable.")
end


function callback_tci(state_u, state_t, rhs, isperiodic, tci::TCI.EntropyProduction)
  @warn "Figure out entropy fluxes for special relativistic euler equation!"
  TODO(callback_tci)
  @unpack cache, mesh                = rhs
  @unpack E, Em1, flx_E, flx_Em1, EP = get_static_variables(cache)
  @unpack flag                       = get_cell_variables(cache)
  @unpack EPjump_lhs, EPjump_rhs     = get_bdry_variables(cache)
  @unpack t, tm1                     = get_global_variables(cache)
  TCI.compute_indicator!(flag, EP, EPjump_lhs, EPjump_rhs,
                         E, Em1, flx_E, flx_Em1, t[1], tm1[1],
                         tci)
end


function callback_tci(state_u, state_t, rhs, isperiodic,
    tci::Union{TCI.Diffusion, TCI.Minmod, TCI.Threshold, 
               TCI.ModalDecayAverage, TCI.ModalDecayHighest})

  @unpack cache, mesh                        = rhs
  @unpack sD, sSr, stau                      = get_dynamic_variables(cache)
  @unpack flag, sD_flag, sSr_flag, stau_flag = get_cell_variables(cache)
  TCI.compute_indicator!(sD_flag, sD, tci)
  TCI.compute_indicator!(sSr_flag, sSr, tci)
  TCI.compute_indicator!(stau_flag, stau, tci)
  _, K = layout(mesh)
  for k = 1:K
    sDf, sSrf, stauf = sD_flag[k], sSr_flag[k], stau_flag[k]
    flag[k] = max(sDf, sSrf, stauf)
  end
  # 'smooth' flag with neighbouring cells by simple averaging
  tmp_flag = sD_flag # reuse memory
  for k = 1:K
    k_lhs, k_rhs = periodic_index(k-1, K), periodic_index(k+1, K)
    if !isperiodic && k == 1
      tmp_flag[k] = (flag[k] + flag[k_rhs]) / 2
    elseif !isperiodic && k == K
      tmp_flag[k] = (flag[k_lhs] + flag[k]) / 2
    else
      tmp_flag[k] = (flag[k_lhs] + flag[k] + flag[k_rhs]) / 3
    end
  end
  @. flag = tmp_flag
end
