function make_callback(rhs::RHS, bdryconds::BoundaryConditions)
  cbfn_equation_step  = invoke_after(rhs, rhs.equation)
  cbfn_equation(u, t) = callback_equation(u, t, rhs, isperiodic(bdryconds), rhs.equation)
  cb_equation         = FunctionCallback(cbfn_equation,
                                         CallbackTiming(every_iteration=1,every_dt=0,
                                                        label="Equation update"))
  cbfn_tci_step       = invoke_after(rhs, rhs.tci)
  cbfn_tci(u, t)      = callback_tci(u, t, rhs, isperiodic(bdryconds), rhs.tci)
  cb_tci              = FunctionCallback(cbfn_tci,
                                         CallbackTiming(every_iteration=1, every_dt=0,
                                                        label="TCI update"))
  cbfn_hrsc_step      = invoke_after(rhs, rhs.hrsc)
  cbfn_hrsc(u, t)     = callback_hrsc(u, t, rhs, isperiodic(bdryconds), rhs.hrsc)
  cb_hrsc = if cbfn_hrsc_step === :substep
    FunctionCallback(cbfn_hrsc, CallbackTiming(force=true,
                                              label="HRSC update"))
  else
    FunctionCallback(cbfn_hrsc, CallbackTiming(every_iteration=1,every_dt=0,
                                               label="HRSC update"))
  end


  cbs_step = [ (cb_equation, cbfn_equation_step),
               (cb_tci, cbfn_tci_step),
               (cb_hrsc, cbfn_hrsc_step) ]
  callbackset_fullstep = CallbackSet()
  callbackset_substep = CallbackSet()
  for (cb, step) in cbs_step
    step === :fullstep && push!(callbackset_fullstep, cb)
    step === :substep && push!(callbackset_substep, cb)
  end

  return callbackset_fullstep, callbackset_substep
end


# as a default run everything after each full time step
invoke_after(rhs, dispatch) = :fullstep


#######################################################################
#                              Equation                               #
#######################################################################


function callback_equation(state_u, state_t, rhs, isperiodic, eq::AbstractScalarEquation)
  @unpack cache, mesh, equation = rhs
  wrap_dynamic_variables!(cache, state_u)
  @unpack u                     = get_dynamic_variables(cache)
  @unpack E, Em1, F, Fm1        = get_static_variables(cache)
  @unpack t, tm1                = get_global_variables(cache)

  # backup previous values
  @. Em1  = E
  @. Fm1  = F
  @. tm1  = t
  t      .= state_t
  # compute new values
  broadcast_volume!(entropy, equation, cache)
  broadcast_volume!(entropy_flux, equation, cache)
end


#######################################################################
#                                HRSC                                 #
#######################################################################


callback_hrsc(state_u, state_t, rhs, isperiodic, hrsc::Nothing) = nothing


callback_hrsc(state_u, state_t, rhs, isperiodic, hrsc::HRSC.AbstractHRSC) =
  error("not yet implemented for type $(typeof(hrsc))")


invoke_after(rhs, hrsc::HRSC.AbstractReconstruction) = :substep
function callback_hrsc(state_u, state_t, rhs, isperiodic,
  hrsc::Union{HRSC.BernsteinReconstruction, HRSC.WENOReconstruction})
  @unpack flag = get_cell_variables(rhs.cache)
  HRSC.reconstruct!(state_u, flag, rhs.hrsc, isperiodic=isperiodic)
end


function callback_hrsc(state_u, state_t, rhs, isperiodic, hrsc::HRSC.AbstractArtificialViscosity)
  @unpack cache, mesh, equation = rhs
  @unpack v         = get_static_variables(cache)
  @unpack mu, max_v = get_cell_variables(cache)
  @unpack u         = get_dynamic_variables(cache)
  broadcast_volume!(speed, equation, cache)
  Npts, K = layout(mesh)
  mat_max_v = reshape(view(v, :), (Npts, K))
  for k = 1:K
    max_v[k] = maximum(view(mat_max_v, :, k))
  end
  HRSC.compute_viscosity!(
      mu,
      u, max_v,
      hrsc)
end


function callback_hrsc(state_u, state_t, rhs, isperiodic, hrsc::HRSC.TCIViscosity)
  @unpack cache, mesh = rhs
  @unpack mu, flag    = get_cell_variables(cache)
  HRSC.compute_viscosity!(mu, flag, hrsc)
end


function callback_hrsc(state_u, state_t, rhs, isperiodic, hrsc::HRSC.SmoothedArtificialViscosity)
  callback_hrsc(state_u, state_t, rhs, isperiodic, hrsc.av)
  @unpack cache, rsolver = rhs
  @unpack mu = get_cell_variables(cache)
  @unpack smoothed_mu = get_static_variables(cache)
  smoothed_mu .= vec(hrsc.smoother(mu,isperiodic))
end


function callback_hrsc(state_u, state_t, rhs, isperiodic, hrsc::HRSC.EntropyViscosity)
  @unpack cache, mesh, equation    = rhs
  @unpack u              = get_dynamic_variables(cache)
  @unpack E, F, Em1, Fm1, v = get_static_variables(cache)
  @unpack mu, max_v      = get_cell_variables(cache)
  @unpack t, tm1         = get_global_variables(cache)
  broadcast_volume!(speed, equation, cache)
  Npts, K = layout(mesh)
  mat_max_v = reshape(view(v, :), (Npts, K))
  for k = 1:K
    max_v[k] = maximum(view(mat_max_v, :, k))
  end
  HRSC.compute_viscosity!(
      mu,
      u, max_v, E, Em1, F, Fm1, t[1], tm1[1],
      hrsc)
end


#######################################################################
#                                 TCI                                 #
#######################################################################


callback_tci(state_u, state_t, rhs, isperiodic, tci::Nothing) = nothing


callback_tci(state_u, state_t, rhs, isperiodic, tci::TCI.AbstractTCI) =
  error("not yet implemented for type $(typeof(tci))")


function callback_tci(state_u, state_t, rhs, isperiodic, tci::TCI.EntropyProduction)
  @unpack cache, mesh            = rhs
  @unpack u                      = get_dynamic_variables(cache)
  @unpack E, Em1, F, Fm1, EP     = get_static_variables(cache)
  @unpack flag                   = get_cell_variables(cache)
  @unpack EPjump_lhs, EPjump_rhs = get_bdry_variables(cache)
  @unpack t, tm1                 = get_global_variables(cache)
  TCI.compute_indicator!(flag, EP, EPjump_lhs, EPjump_rhs,
                         E, Em1, F, Fm1, t[1], tm1[1],
                         tci)
end


function callback_tci(state_u, state_t, rhs, isperiodic,
    tci::Union{TCI.Diffusion, TCI.Minmod, TCI.Threshold,
               TCI.ModalDecayAverage, TCI.ModalDecayHighest})
  @unpack cache, mesh            = rhs
  @unpack u                      = get_dynamic_variables(cache)
  @unpack flag                   = get_cell_variables(cache)
  TCI.compute_indicator!(flag, u, tci)
end
