"""
    abstract type AbstractScalarEquation <: dg1d.AbstractEquation end

Supertype for ScalarEquation problems.
New problems should subtype it and also implement the following methods:
- `make_Equation(prms, ::Type{NewProblem})`: factory for an instance of type `NewProblem`.
The following methods have to be defined with the `@with_signature` interface:
- `speed(args, ::NewProblem)`
- `flux(args, ::NewProblem)`
- `entropy(args, ::NewProblem)`
- `entropy_flux(args, ::NewProblem)`
"""
abstract type AbstractScalarEquation <: dg1d.AbstractEquation end


#######################################################################
#                          Equation Factory                           #
#######################################################################


function make_Equation(prms)
  available_types = concretetypes(AbstractScalarEquation)
  all_options = last.(split.(string.(available_types), "."))
  type = query(prms, "type", options=all_options)
  expr = quote $(Symbol(type)) end
  return make_Equation(prms, eval(expr))
end


#######################################################################
#                          Advection                          #
#######################################################################

struct Advection <: AbstractScalarEquation
  v::Float64
end

function make_Equation(prms, ::Type{Advection})
  v = query(Float64, prms, "v", default=1.0)
  return Advection(v)
end

@with_signature function flux(eq::Advection) 
  @accepts State(u)
  flx_u = eq.v * u
  @returns flx_u
end

@with_signature function speed(eq::Advection)
  @accepts State(u)
  v = eq.v
  @returns v
end

@with_signature function entropy(eq::Advection)
  @accepts u
  E = u^2 / 2
  @returns E
end

@with_signature function entropy_flux(eq::Advection)
  @accepts u
  F = u^2 / 2
  @returns F
end

#######################################################################
#                           Burgers                           #
#######################################################################

struct Burgers <: AbstractScalarEquation end

make_Equation(prms, ::Type{Burgers}) = Burgers()

make_Burgers(prms) = Burgers()

entropy!(E, u, eq::Burgers, mesh::Mesh) = @. E = u^2 / 2

entropyflux!(F, u, eq::Burgers, mesh::Mesh) = @. F = u^3 * 2 / 3

@with_signature function flux(eq::Burgers) 
  @accepts State(u)
  flx_u = u^2 / 2
  @returns flx_u
end

@with_signature function speed(eq::Burgers)
  @accepts State(u)
  v = u
  @returns v
end

@with_signature function entropy(eq::Burgers)
  @accepts u
  E = u^2 / 2
  @returns E
end

@with_signature function entropy_flux(eq::Burgers)
  @accepts u
  F = u^3 * 2 / 3
  @returns F
end

#######################################################################
#                                 LDG                                 #
#######################################################################

@with_signature function ldg_flux(eq::AbstractScalarEquation)
  @accepts State(u)
  flx_q = u
  @returns flx_q
end

@with_signature function ldg_speed(eq::AbstractScalarEquation)
  @accepts State(u)
  v_q = 1.0
  @returns v_q
end

@with_signature function av_flux(eq::AbstractScalarEquation)
  @accepts State(q), smoothed_mu
  flx_q = smoothed_mu * q
  @returns flx_q
end

@with_signature function av_speed(eq::AbstractScalarEquation)
  @accepts State(q), smoothed_mu
  v_av_q = smoothed_mu
  @returns v_av_q
end
