Base.@kwdef struct ModalDecayAverage <: AbstractTCI
  mesh
  # parameters
  threshold_min::Float64
  threshold_max::Float64
end


# TODO Issue warning that this is expected to work properly only for N > 4 or so.
function parse_parameters(::Type{ModalDecayAverage}, prms)
  threshold_min = query(Float64, prms, "threshold_min", default=1.0)
  threshold_max = query(Float64, prms, "threshold_max", default=3.0)
  @assert threshold_min > 0
  @assert threshold_max > 0
  @assert threshold_min < threshold_max
  return (; threshold_min, threshold_max)
end


function compute_indicator!(
    flag,   # mutated outputs
    u,      # inputs
    tci::ModalDecayAverage)

  @unpack mesh       = tci
  @unpack invjac, dl = mesh
  @unpack D, invV    = mesh.element

  L       = layout(mesh)
  Npts, K = L
  N       = Npts - 1
  u       = reshape(u, L)

  if Npts <= 2
    error("ModalDecayAverage TCI is unrelieable for Npts <= 2")
  end

  # perfect decay modal coefficients, added to enhance sensitivity to 
  # floating point noise
  vhat = collect(Float64, 1:N).^(-N)
  vhat ./= norm(vhat)

  cellnorm2_u = dg1d.cellwise_inner_product(u, u, mesh)
  cellnorm2_u[abs.(cellnorm2_u) .< 1e-10 ] .= 1.0

  # estimate effective decay rate by least square fit of modal coefficients, (3.9)
  # sum_m ( log(|q_m|) - (log(c) - tau log(m)) )^2 = sum_m (y_m - (a + b x_m))^2 -> min
  # log(m) ... grid points in log scale
  rng_N = 1:N
  data_x = @. - log(rng_N)
  uhat = zeros(Npts)

  @unpack threshold_min, threshold_max = tci
  for k in 1:K
    uhat .= invV * u[:,k]
    # (3.11)
    # only consider coefficients 2:Np (which corresponds to 1:N), because
    # the 1st (0th) one corresponds to a constant mode that is irrelevant
    # for the smoothness esimation.
    mod_uh = @. sqrt( uhat[2:end]^2 + cellnorm2_u[k] * vhat^2 )
    # (3.12)
    skylined_mod_uh = [ maximum(mod_uh[min(n,N-1):end]) for n in 1:N ]
    # skylined_mod_uh = zeros(size(mod_uh))
    # for m in 1:N
    #   skylined_mod_uh[m] = maximum(mod_uh[min(m,N-1):end])
    # end

    # data points in log scale
    data_y = @. log(skylined_mod_uh)
    data_y[isinf.(data_y)] .= -20.0
    _, _, tau = dg1d.least_sqr_linear(data_x, data_y)
    isnan(tau) && error("undesired nan encountered")

    if threshold_max < tau
      flag[k] = 0.0
    elseif threshold_min <= tau <= threshold_max
      flag[k] = 1.0 - (tau - threshold_min) / (threshold_max - threshold_min)
    else
      flag[k] = 1.0
    end
  end

  return
end
