@testset "Utilities" begin


@testset "Least squares fit" begin

  npts = 57
  a = 5.7
  b = -12.4
  x = collect(Float64, 1:npts)
  y = @. a + b * x

  f_fit, a_fit, b_fit = dg1d.least_sqr_linear(x, y)

  @test a_fit ≈ a
  @test b_fit ≈ b

  # test with random data point
  i = Int(rand(x))
  @test f_fit(x[i]) ≈ y[i]

end


@testset "FD differences" begin

  N = 2000
  L = 2.0
  dx = L / (N-1) # full step width

  # generate grid with randomly sampled grid spacings within range [0, L]
  x = [abs(randn())*dx]
  while true
    rnd = abs(randn())
    xi = dx * rnd + x[end]
    if xi > L
      break
    end
    push!(x, xi)
  end

  # test function
  f = @. x^2 - 2.0 * exp(-x^2)
  fx = @. 2.0 * x + 4.0 * exp(-x^2) * x
  fxx = @. 2.0 - 8.0 * exp(-x^2) * x^2 + 4.0 * exp(-x^2)

  nfx, nfxx = dg1d.fd_approx(f, x)
  xx = x[2:end-1]
  fx = fx[2:end-1]
  fxx = fxx[2:end-1]

  # think about reasonable numerical error to test against here
  # @test fx[2:end-1] ≈ nfx
  # @test fxx[2:end-1] ≈ nfxx

  # Plots.gr()
  # plt = Plots.plot()
  # Plots.plot!(plt, xx, fx - nfx, label="f_x (ana vs num)")
  # Plots.plot!(plt, xx, fxx - nfxx, label="f_xx (ana vs num)")

  # output_dir = dg1d.get_output_dir()
  # filename = "test_fd_approximation.pdf"
  # Plots.savefig(plt, joinpath(output_dir, filename))
  # @info "Test figure saved to $filename ..."

end


@testset "Numerical utilities" begin

  # discrete case
  N = 3
  index = [ 1, 2, 3, 4, 5 ]
  obtained_index = periodize.(index, N)
  expected_index = [ 1, 2, 3, 1, 2 ]
  @test obtained_index == expected_index
  @test_throws Exception periodize.(index, 0)

  # continuum case
  L = convert(Float64, N)
  coordinate = convert.(Float64, index)
  obtained_coordinate = periodize.(coordinate, L)
  expected_coordinate = convert.(Float64, expected_index)
  @test all(isapprox.(obtained_coordinate, expected_coordinate))
  @test all(isnan.(periodize.(coordinate, 0.0)))

end


@testset "TOML expression string" begin

  expr_str = "\$(x = 1)" # need to escape here, but not in TOML file
  obtained = dg1d.Sandbox.parse_expression_string(expr_str)
  expected = 1
  @test obtained == expected

  expr_str = "\$(47)" # need to escape here, but not in TOML file
  obtained = dg1d.Sandbox.parse_expression_string(expr_str)
  expected = 47
  @test obtained == expected

  expr_str = "\$(x = 1; y = x + 1)" # need to escape here, but not in TOML file
  obtained = dg1d.Sandbox.parse_expression_string(expr_str)
  expected = 2
  @test obtained == expected

  expr_str = """\$(
    x = 1
    y = x + 2
    z = y + 3
    )""" # need to escape here, but not in TOML file
  obtained = dg1d.Sandbox.parse_expression_string(expr_str)
  expected = 6
  @test obtained == expected

  # 'simple' functions are allowed
  expr_str = "\$(collect(1:5))" # need to escape here, but not in TOML file
  obtained = dg1d.Sandbox.parse_expression_string(expr_str)
  expected = collect(1:5)
  @test obtained == expected

  # powerful tokens are blacklisted
  @test_throws ErrorException dg1d.Sandbox.parse_expression_string("\$(import LinearAlgebra)")
  @test_throws ErrorException dg1d.Sandbox.parse_expression_string("\$(using LinearAlgebra)")
  @test_throws ErrorException dg1d.Sandbox.parse_expression_string(
    """\$(
    function hello_world()
      println("Hello")
    end
    )""")
  @test_throws ErrorException dg1d.Sandbox.parse_expression_string(
    """\$(
    if 1 > 0
      return 1
    else
      return 0
    end
    )""")

  # errornous expression string, because uvw not defined
  @test_throws ErrorException dg1d.Sandbox.parse_expression_string("\$(uvw - 1)")
  
  # watch out: previously evaluated expressions remain inside dg1d.Sandbox
  dg1d.Sandbox.parse_expression_string("\$(x = 123)") # dg1d.Sandbox.x = 123
  obtained = dg1d.Sandbox.parse_expression_string("\$(y = x + 1)") # dg1d.Sandbox.y = 124
  expected = 124
  @test obtained == expected

  # returns initial string if it is not of the form $( ... )
  obtained = dg1d.Sandbox.parse_expression_string("non-expression string") # dg1d.Sandbox.y = 124
  expected = "non-expression string"
  @test obtained == expected

  # also returns initial string if pattern is incomplete
  obtained = dg1d.Sandbox.parse_expression_string("\$( x = 1 + 12")
  expected = "\$( x = 1 + 12"
  @test obtained == expected

end


@testset "Query parameters" begin

  prms = Dict("int"=>1, "str"=>"String", "float"=>Float64(π), "vector"=>[1, 2, 3])

  # standard query
  int = query(prms, "int")
  @test int == prms["int"]
  int = query(prms, "int", default=4) # use default, has no effect becase 'int' is present
  @test int == prms["int"]
  str = query(prms, "str")
  @test str == prms["str"]
  float = query(prms, "float")
  @test isapprox(float, prms["float"])
  vector = query(prms, "vector")
  @test isapprox(vector, prms["vector"])

  # query specific type
  int = query(Int, prms, "int")
  @test int == prms["int"]
  int = query(Int, prms, "int") # down conversions work too
  @test int == prms["int"]
  str = query(String, prms, "str")
  @test str == prms["str"]
  float = query(Number, prms, "float") # Number is a very loose type
  @test isapprox(float, prms["float"])
  vector = query(Vector{Int64}, prms, "vector") # Number is a very loose type
  @test isapprox(vector, prms["vector"])
  
  # query with defaults
  int = query(prms, "int", default=5)
  @test int == prms["int"]
  int = query(prms, "myint", default=5)
  @test int == 5
  str = query(prms, "str", default="abc")
  @test str == prms["str"]
  str = query(prms, "mystr", default="abc")
  @test str == "abc"
  float = query(prms, "float", default=1.23)
  @test float == prms["float"]
  float = query(prms, "myfloat", default=1.23)
  @test float == 1.23
  vector = query(prms, "myvector", default=[3,4,5])
  @test vector == [3,4,5]

  # query with options
  int = query(prms, "int", options=[1,2,3])
  @test int == prms["int"]
  str = query(prms, "str", options=["Option1", "String", "Option3"])
  @test str == prms["str"]
  # you can also match against parts of a string
  #   1. first occurence in options vector 
  str = query(prms, "str", options=["ring", "Stri", "String"])
  str = query(prms, "str", options=["String", "ring", "Stri"])
  @test str == "String" # matches because String comes before ring
  float = query(prms, "float", options=Float64[π+0.1, π, π-0.1])
  @test isapprox(float, prms["float"])
  vector = query(prms, "vector", options=[ [1,3], [1,2,3], [3,4,5,7] ])
  @test isapprox(vector, prms["vector"])


  # test failures
  @test_throws ErrorException query(prms, "int", options=[2,3]) # prms["int"] = 1
  @test_throws ErrorException query(prms, "str", options=["Option1", "Option2"]) # prms["str"] = "String"
  @test_throws ErrorException query(prms, "float", options=Float64[π-0.1,π+0.1]) # prms["float"] = π
  @test_throws ErrorException query(prms, "vector", options=["Option1", 1, 3, 4]) # prms["vector"] = [1,2,3]

  # conversion where value is of wrong type
  @test_throws ErrorException query(String, prms, "int")
  @test_throws ErrorException query(Float64, prms, "str")
  # note that conversions from AbstractFloat to Integer employs rounding
  @test query(Int, prms, "float") == 3 # prms["float"] = π ≈ 3
  # conversion from Integer to AbstractFloat works too
  @test query(Float64, prms, "int") == 1.0 # prms["int"] = 1 ≈ 1.0
  # conversion in roundings also applies when reading vectors
  @test query(Vector{Float64}, prms, "vector") == [1.0, 2.0, 3.0] # prms["int"] = 1 ≈ 1.0

  # conversions where default is of wrong type
  @test_throws ErrorException query(String, prms, "myint", default=1)
  @test_throws ErrorException query(Float64, prms, "mystr", default="str")
  @test_throws ErrorException query(Int, prms, "mystr", default="str")
  @test_throws ErrorException query(Int, prms, "myvector", default=["abc", "def", "ghi"])

	# duplicated options
  prms = Dict("abc"=>1)
	@test_throws ErrorException query(String, prms, "abc", options=[1,1])
	@test_throws ErrorException query(String, prms, "abc", options=[[1],[1]])
	@test_throws ErrorException query(String, prms, "abc", options=["def","def"])

  ### query with regex
  prms = Dict("abc"=>1, "abc.def"=>2, "def.def"=>3, "def.abc"=>4)
  
  obtained = query(prms, r"abc")
  expected = Dict("abc"=>1, "abc.def"=>2, "def.abc"=>4)
  @test obtained == expected

  obtained = query(prms, r"def")
  expected = Dict("abc.def"=>2, "def.def"=>3, "def.abc"=>4)
  @test obtained == expected

  obtained = query(prms, r"^abc")
  expected = Dict("abc"=>1, "abc.def"=>2)
  @test obtained == expected

  obtained = query(prms, r"\.abc")
  expected = Dict("def.abc"=>4)
  @test obtained == expected

  obtained = query(prms, r"(abc)|(def)")
  expected = prms
  @test obtained == expected

  # issue #59
	prms = Dict("type"=>"bugner_gaussian")
	obtained = query(String, prms, "type")
	expected = "bugner_gaussian"
	@test obtained == expected
	@test_throws ErrorException query(String, prms, "type", options=["gaussian"])

  # parse expression strings
  prms = Dict("list"=>"\$(collect(1:5))") # escaping of $ only need here, but no in TOML files
  obtained = query(prms, "list")
  expected = collect(1:5)
  @test obtained == expected

  # errornous expression strings throw
  prms = Dict("list"=>"\$(colct(1:5))") # escaping of $ only need here, but no in TOML files
  @test_throws ErrorException query(prms, "list")

  # incomplete expression strings just return the string
  prms = Dict("result"=>"\$(47 + 1.4") # missing brace at the end
  obtained = query(prms, "result")
  expected = "\$(47 + 1.4"
  @test obtained == expected

end


@testset "Dictionary operations" begin

  d1 = Dict("a"=>1, "b"=>2, "c"=>"abc")
  d2 = Dict("a"=>2, "b"=>"f", "c"=>"abc", "d"=>15)

  amend!(d1, d2)
  @test d1["a"] == 1
  @test d1["b"] == 2
  @test d1["c"] == "abc"
  @test d1["d"] == 15
  @test samekeys(d1,d2) == true
  @test diffkeys(d1,d2) == []

  d1["e"] = "element with no defaults in d2"
  amend!(d1, d2)
  @test d1["e"] == d1["e"]
  @test samekeys(d1,d2) == false
  @test diffkeys(d1,d2) == ["e"]

  ## flatten dicts
  nested_dict = Dict("d1"=>Dict("a"=>1, "b"=>2),
                     "d2"=>Dict("c"=>3, "d"=>4))
  flattened_dict = dg1d.flatten(nested_dict)
  expected = Dict( "d1.a" => 1, "d1.b" => 2, "d2.c" => 3, "d2.d" => 4)
  @test flattened_dict == expected

  ## trim dict keys
  d = Dict("abc"=>1, "abc.def"=>2, "abcdef"=>3)
  obtained = trim_keys(d, "a")
  expected = Dict("bc"=>1, "bc.def"=>2, "bcdef"=>3)
  @test obtained == expected

  obtained = trim_keys(d, "ab")
  expected = Dict("c"=>1, "c.def"=>2, "cdef"=>3)
  @test obtained == expected

  @test_throws ErrorException trim_keys(d, "abc")
  @test_throws ErrorException trim_keys(d, "cba") # wrong pattern

  @test trim_keys(Dict(), "") == Dict() # nothing to trim

  obtained = trim_keys(d, "")
  expected = d
  @test obtained == expected

  ## regex match dict keys
  # return matched and unmatched in separate dictionaries
  mesh_prms = Dict("Mesh.xrange" => [0, 1],
                   "Me_sh.N"     => 4,
                   "Mesh__CFL"   => 0.1,
                   "__Mesh.K"    => 40)
  obtained_matched, obtained_unmatched = match_keys(k->match(r"Mesh\.", k), mesh_prms)
  expected_matched = Dict("Mesh.xrange"   => [0,1], 
                          "__Mesh.K"      => 40) # because regex does not start with ^
  expected_unmatched = Dict("Me_sh.N"     => 4, 
                            "Mesh__CFL"   => 0.1)
  @test obtained_matched == expected_matched
  @test obtained_unmatched == expected_unmatched

  ## prepend dict keys
  d = Dict("abc"=>1, "def"=>2, "ghi"=>3)
  obtained = prepend_keys(d, "xyz")
  expected = Dict("xyz.abc"=>1, "xyz.def"=>2, "xyz.ghi"=>3)
  @test obtained == expected

  ## Convert DataFrame rows to dictionaries
  df = DataFrame(A=["a1", "a2", "a3"],
                 B=["b1", "b2", "b3"],
                 C=["c1", "c2", "c3"])
  expected_dicts = [ Dict("A"=>"a1", "B"=>"b1", "C"=>"c1"),
                     Dict("A"=>"a2", "B"=>"b2", "C"=>"c2"),
                     Dict("A"=>"a3", "B"=>"b3", "C"=>"c3") ]
  for (row, expected) in zip(eachrow(df), expected_dicts)
    obtained = to_dict(row)
    @test obtained == expected
    obtained_flatdict = to_dict(FlatDict, row)
    expected_flatdict = FlatDict(expected)
    @test obtained_flatdict == expected_flatdict
  end
  
end


@testset "String operations" begin

  dummypath = "/this/is/a/path/string.with.lots.of.extensions"
  filename, exts = splitallext(dummypath)
  @test filename == "/this/is/a/path/string"
  @test exts == ".with.lots.of.extensions"

end


@testset "Filesystem operations" begin

  # my version of normpath
  # always add a trailing space to path name of a directory

  thisdir = Filesystem.abspath(@__DIR__)
  mockdir = Filesystem.abspath(joinpath(thisdir, "../mockdir"))
  ispath(mockdir) && rm(mockdir, recursive=true)
  mkdir(mockdir)

  # mock parfiles and output dirs
  dummyoutputdir = Filesystem.abspath(joinpath(mockdir, "testparfile"))
  isdir(dummyoutputdir) && rm(dummyoutputdir, recursive=true)
  mkdir(dummyoutputdir)

  # running file at mockdir/testparfile.toml should generate folder mockdir/testparfile
  dummyparfile1 = joinpath(mockdir, "testparfile.some.weird.extension.toml")
  touch(dummyparfile1)
  # running file at testparfile/testparfile.toml should backup folder testparfile/ as
  # testparfile.previous/ and then generate folder mockdir/testparfile
  dummyparfile2 = joinpath(dummyoutputdir, "testparfile.toml")
  touch(dummyparfile2)

  dummyparfiles = [ dummyparfile1, dummyparfile2 ]

  for dp in dummyparfiles
    path = mk_outputdir(dp)
    @test isdir(path)
    # call to already existing should generate a backup at $path.previous
    path = mk_outputdir(dp)
    path_previous = "$(path).previous"
    @test isdir(path)
    @test isdir(path_previous)
  end

  nofilethere = joinpath(mockdir, "nofilethere.toml")
  @test_throws ErrorException mk_outputdir(nofilethere)

  wrongextensionfile = joinpath(mockdir, "wrongextension.toml.blabla")
  touch(wrongextensionfile)
  @test_throws ErrorException mk_outputdir(wrongextensionfile)


  #### mock a project directory
  projectdirname = "projectdir"
  projectdir = joinpath(mockdir, projectdirname)
  mkdir(projectdir)
  # add some parameter files
  projects = [joinpath(projectdir, "$(projectdirname)_$i") for i in 1:5]
  mkdir.(projects)
  projecttomls = [ "$(dir).toml" for dir in projects ]
  touch.(projecttomls)
  # fill project folders with toml files
  subprojecttomls = [ joinpath(dir, "$(projectdirname)_$i.toml") for (i, dir) in enumerate(projects) ]
  touch.(subprojecttomls)

  ### test getters and gather methods
  obtained_projectdirname = dirname_path(projecttomls[1])
  @test obtained_projectdirname == projectdirname
  obtained_projectdirname = dirname_path(projectdir)
  @test obtained_projectdirname == projectdirname

  # basename and dirname behave differently depending on whether the path has a trailing / or not
  # dirname_path should not depend on that and it should always return the dirname of the 
  # deepest nested directory
  testdir_dirname_path = joinpath(mockdir, "test_dirname_path")
  mkdir(joinpath(testdir_dirname_path))
  path_noslash = testdir_dirname_path
  path_slash = "$(testdir_dirname_path)/" # will break on windows, but who cares?
  dirname_noslash = dirname_path(path_noslash)
  dirname_slash = dirname_path(path_slash)
  @test dirname_noslash == dirname_slash
  # if path is a file, then return the dirname of that file
  testfile_dirname_path = joinpath(testdir_dirname_path, "testfile_dirname_path")
  touch(testfile_dirname_path)
  @test dirname_path(testfile_dirname_path) == "test_dirname_path"
  # throw if path does not exist
  @test_throws ErrorException dirname_path(tempname())

  obtained_parameter_file = get_parameter_file(projects[1])
  @test obtained_parameter_file == Filesystem.basename(subprojecttomls[1])

  gathered_projects = gather_output_directories(projectdir)
  @test gathered_projects == projects

  # write some parameters to files
  # need two different sets to test duplic
  parameters1 = Dict("a"=>1, "b"=>2, "c"=>"hello world", "d"=>[1,2,3])
  parameters2 = Dict("e"=>123)
  parameters3 = Dict("xyz"=>"uno dos tres")
  function write_parameters(toml, ps)
    open(toml, "w") do io
      TOML.print(io, ps)
    end
  end
  write_parameters.(projecttomls[1:2], Ref(parameters1))
  write_parameters.(projecttomls[3:4], Ref(parameters2))
  write_parameters(projecttomls[5], parameters3)

  obtained_duplicates = find_duplicated_parameter_files(projectdir)
  expected_duplicates = [ [ projecttomls[1:2]... ], [ projecttomls[3:4]... ] ]
  @test obtained_duplicates == expected_duplicates

  df_parameters = gather_parameters(projectdir)
  @info "Parsed parameters from project directory '$projectdir'"
  display(df_parameters)

  # clean up mocking
  isdir(mockdir) && rm(mockdir, recursive=true)

end


@testset "DataFrame helpers" begin
  
  testdata = joinpath(dirname(@__FILE__), "../testdata") |> normpath |> abspath
  table = gather_parameters(testdata)

  # match against one argument with a single option
  with_properties = Dict("Mesh.N" => 4) # only one option
  subtable, diffs = match_properties(table, with_properties)
  @test all(subtable."Mesh.N" .== with_properties["Mesh.N"])
  @test haskey(diffs, "Evolution.algorithm")
  @test haskey(diffs, "filename")
  @test haskey(diffs, "Mesh.K")

  tmp_subtable, tmp_diffs = copy(subtable), copy(diffs)
  with_properties = Dict("Mesh.N" => [4]) # same as 4
  subtable, diffs = match_properties(table, with_properties)
  @test tmp_subtable == subtable
  @test tmp_diffs == diffs

  # match against one arugment with multiple options
  with_properties = Dict("Mesh.N" => 4, "Mesh.K" => [30, 40, 50, 60])
  subtable, diffs = match_properties(table, with_properties)
  # there are not files that match "Mesh.N" = 4 and "Mesh.K" = 60
  @test all(subtable."Mesh.N" .== 4)
  @test !(60 in diffs["Mesh.K"])

  # "Evolution.algorithm" is either "RK4" or "RKF10"
  @test issubset(diffs["Evolution.algorithm"], [ "RK4", "RKF10" ])

  # extract "RK4" files
  with_properties = Dict("Evolution.algorithm" => "RK4")
  rk4_subtable, rk4_diffs = match_properties(subtable, with_properties)
  @test !haskey(rk4_diffs, "Evolution.algorithm")

  # ignore column 'filename' when matching
  subtable, diffs = match_properties(table, with_properties, ignore=["filename"])
  @test !haskey(diffs, "filename")

end


end
